#
# mercurial - a minimal scalable distributed SCM
# v0.1
#
# Copyright 2005 Matt Mackall <mpm@selenic.com>
#
# This software may be used and distributed according to the terms
# of the GNU General Public License, incorporated herein by reference.

import sys, zlib, struct, mdiff, sha, socket
import binascii, os, tempfile, time, base64, re

nullid = sha.sha().digest()

class revlog:
    def __init__(self, indexfile, datafile):
        self.indexfile = indexfile
        self.datafile = datafile
        self.index = []
        self.nodemap = { -1: nullid, nullid: -1 }
        try:
            n = 0
            i = self.open(self.indexfile).read()
            for f in range(0, len(i), 40):
                # offset, size, base, p1, p2, nodeid
                e = struct.unpack(">5l20s", i[f:f+40])
                self.nodemap[e[5]] = n
                self.index.append(e)
                n += 1
        except IOError: pass

    def open(self, fn, mode="r"): return file(fn, mode)

    def tip(self): return len(self.index) - 1
    def node(self, rev): return rev < 0 and nullid or self.index[rev][5]
    def rev(self, node): return self.nodemap[node]

    def parents(self, rev): return self.index[rev][3:5]
    def start(self, rev): return self.index[rev][0]
    def length(self, rev): return self.index[rev][1]
    def end(self, rev): return self.start(rev) + self.length(rev)
    def base(self, rev): return self.index[rev][2]

    def ancestor(self, a, b):
        def expand(e1, e2, a1, a2):
            ne = []
            for r in e1:
                (p1, p2) = self.parents(r)
                if p1 in a2: return p1
                if p2 in a2: return p2
                if p1 not in a1:
                    a1[p1] = 1
                    ne.append(p1)
                    if p2 >= 0 and p2 not in a1:
                        a1[p2] = 1
                        ne.append(p2)
            return expand(e2, ne, a2, a1)
        return expand([a], [b], {a:1}, {b:1})

    def mergedag(self, other, accumulate = None):
        amap = self.nodemap
        bmap = other.nodemap
        old = i = self.tip()
        l = []

        for r in range(other.tip() + 1):
            id = other.node(r)
            if id not in amap:
                i += 1
                (x, y) = other.parents(r)
                xn, yn = other.node(x), other.node(y)
                l.append((r, amap[xn], amap[yn]))
                amap[id] = i

        r = other.revisions([e[0] for e in l])
        for e in l:
            t = r.next()
            if accumulate: accumulate(t)
            self.addrevision(t, e[1], e[2])

        return (old, self.tip())

    def resolvedag(self, old, new):
        if old == new: return None
        a = self.ancestor(old, new)
        print old, new, a
        if old == a: return new
        return self.merge3(old, new, a)

    def merge(self, other):
        # this is where to handle rollback
        (o, n) = self.mergedag(other)
        return self.resolvedag(o, n)

    def revisions(self, list):
        # this can be optimized to do spans, etc
        # be stupid for now
        for r in list:
            yield self.revision(r)

    def revision(self, rev):
        if rev == -1: return ""
        
        base = self.base(rev)
        start = self.start(base)
        end = self.end(rev)

        f = self.open(self.datafile)
        f.seek(start)
        data = f.read(end - start)

        last = self.length(base)
        text = zlib.decompress(data[:last])

        for r in range(base + 1, rev + 1):
            s = self.length(r)
            b = zlib.decompress(data[last:last + s])
            text = mdiff.patch(text, b)
            last = last + s

        (p1, p2) = self.parents(rev)
        n1, n2 = self.node(p1), self.node(p2)
        node = sha.sha(n1 + n2 + text).digest()
        if self.node(rev) != node:
            raise "Consistency check failed on %s:%d" % (self.datafile, rev)

        return text  

    def addrevision(self, text, p1=None, p2=None):
        if text is None: text = ""
        if p1 is None: p1 = self.tip()
        if p2 is None: p2 = -1

        t = self.tip()
        n = t + 1

        if n:
            start = self.start(self.base(t))
            end = self.end(t)
            prev = self.revision(t)
            data = zlib.compress(mdiff.textdiff(prev, text))

        # full versions are inserted when the needed deltas
        # become comparable to the uncompressed text
        if not n or (end + len(data) - start) > len(text) * 2:
            data = zlib.compress(text)
            base = n
        else:
            base = self.base(t)

        offset = 0
        if t >= 0:
            offset = self.end(t)

        n1, n2 = self.node(p1), self.node(p2)
        node = sha.sha(n1 + n2 + text).digest()
        e = (offset, len(data), base, p1, p2, node)
        
        self.index.append(e)
        entry = struct.pack(">5l20s", *e)
        self.nodemap[node] = n
        self.open(self.indexfile, "a").write(entry)
        self.open(self.datafile, "a").write(data)
        return n

    def merge3(self, my, other, base):
        def temp(prefix, rev):
            (fd, name) = tempfile.mkstemp(prefix)
            f = os.fdopen(fd, "w")
            f.write(self.revision(my))
            os.close(fd)
            return name
        a = temp("local", my)
        b = temp("remote", other)
        c = temp("parent", base)
        # call out to merge here, return success flag
        cmd = os.environ["HGMERGE"]
        r = os.system("%s %s %s %s" % (cmd, a, b, c))
        if r: raise "Merge failed, implement rollback!"
        t = open(a).read()
        return self.addrevision(t, my, other)

class filelog(revlog):
    def __init__(self, repo, path):
        self.repo = repo
        s = sha.sha(path).digest()
        s = base64.encodestring(s)[:-3]
        s = re.sub("\+", "%", s)
        s = re.sub("/", "_", s)
        revlog.__init__(self, os.path.join("index/", s),
                        os.path.join("data/", s))
    def open(self, file, mode = "r"):
        return self.repo.open(file, mode)

class manifest(revlog):
    def __init__(self, repo):
        self.repo = repo
        revlog.__init__(self, "00manifest.i", "00manifest.d")
    def open(self, file, mode="r"):
        return self.repo.open(file, mode)
    def manifest(self, rev):
        text = self.revision(rev)
        map = {}
        for l in mdiff.linesplit(text):
            map[l[41:-1]] = binascii.unhexlify(l[0:40])
        return map
    def addmanifest(self, map, p1=None, p2=None):
        files = map.keys()
        files.sort()
        text = "".join(["%s %s\n" % (binascii.hexlify(map[f]), f)
                        for f in files])
        return self.addrevision(text, p1, p2)

class changelog(revlog):
    def __init__(self, repo):
        self.repo = repo
        revlog.__init__(self, "00changelog.i", "00changelog.d")
    def open(self, file, mode="r"):
        return self.repo.open(file, mode)

    def extract(self, text):
        l = mdiff.linesplit(text)
        manifest = binascii.unhexlify(l[0][:-1])
        user = l[1][:-1]
        date = l[2][:-1]
        last = l.index("\n")
        files = [f[:-1] for f in l[3:last]]
        desc = "".join(l[last+1:])
        return (manifest, user, date, files, desc)

    def changeset(self, rev):
        return self.extract(self.revision(rev))
        
    def addchangeset(self, manifest, list, desc, p1=None, p2=None):
        try: user = os.environ["HGUSER"]
        except: user = os.environ["LOGNAME"] + '@' + socket.getfqdn()
        date = "%d %d" % (time.time(), time.timezone)
        list.sort()
        l = [binascii.hexlify(manifest), user, date] + list + ["", desc]
        text = "".join([e + "\n" for e in l])
        return self.addrevision(text, p1, p2)

    def merge3(self, my, other, base):
        pass

class repository:
    def __init__(self, path=None, create=0):

        if not path:
            p = os.getcwd()
            while not os.path.isdir(os.path.join(p, ".hg")):
                p = os.path.dirname(p)
                if p == "/": raise "No repo found"
            path = p

        self.root = path
        self.path = os.path.join(path, ".hg")

        if create:
            os.mkdir(self.path)  
            os.mkdir(os.path.join(self.path, "data"))
            os.mkdir(os.path.join(self.path, "index"))

        self.manifest = manifest(self)
        self.changelog = changelog(self)

        try:
            self.current = int(self.open("current").read())
        except:
            self.current = None
      
    def open(self, path, mode="r"):
        f = self.join(path)
        if mode == "a" and os.path.isfile(f):
            s = os.stat(f)
            if s.st_nlink > 1:
                file(f+".tmp", "w").write(file(f).read())
                os.rename(f+".tmp", f)
        
        return file(self.join(path), mode)

    def join(self, f):
        return os.path.join(self.path, f)

    def file(self, f):
        return filelog(self, f)
            
    def merge(self, other):
        changed = {}
        new = {}
        def accumulate(text):
            files = self.changelog.extract(text)[3]
            for f in files:
                print " ",f,"changed"
                changed[f] = 1

        # begin the import/merge of changesets
        print "beginning changeset merge"
        (co, cn) = self.changelog.mergedag(other.changelog, accumulate)

        if co == cn: return

        # merge all files changed by the changesets,
        # keeping track of the new tips
        changed = changed.keys()
        changed.sort()
        for f in changed:
            print "merging", f
            f1 = filelog(self, f)
            f2 = filelog(other, f)
            rev = f1.merge(f2)
            if rev:
                new[f] = f1.node(rev)

        # begin the merge of the manifest
        print "merging manifests"
        (mm, mo) = self.manifest.mergedag(other.manifest)
        ma = self.manifest.ancestor(mm, mo)

        # resolve the manifest to point to all the merged files
        print "resolving manifests"
        mmap = self.manifest.manifest(mm) # mine
        omap = self.manifest.manifest(mo) # other
        amap = self.manifest.manifest(ma) # ancestor
        nmap = {}

        for f, mid in mmap.items():
            if f in omap:
                if mid != omap[f]: 
                    nmap[f] = new.get(f, mid) # use merged version
                else:
                    nmap[f] = new.get(f, mid) # they're the same
                del omap[f]
            elif f in amap:
                if mid != amap[f]: 
                    pass # we should prompt here
                else:
                    pass # other deleted it
            else:
                nmap[f] = new.get(f, mid) # we created it

        del mmap

        for f, oid in omap.items():
            if f in amap:
                if oid != amap[f]:
                    pass # this is the nasty case, we should prompt here too
                else:
                    pass # probably safe
            else:
                nmap[f] = new.get(f, mid) # remote created it

        del omap
        del amap

        nm = self.manifest.addmanifest(nmap, mm, mo)
        node = self.manifest.node(nm)

        # Now all files and manifests are merged, we add the changed files
        # and manifest id to the changelog
        print "committing merge changeset"
        new = new.keys()
        new.sort()
        if co == cn: cn = -1
        # we should give the user an opportunity to edit the changelog desc
        self.changelog.addchangeset(node, new, "merge", co, cn)

    def commit(self):
        try:
            update = [ l[:-1] for l in self.open("to-add") ]
        except IOError:
            update = []
        try:
            delete = [ l[:-1] for l in self.open("to-delete") ]
        except IOError:
            delete = []

        # check in files
        new = {}
        for f in update:
            r = filelog(self, f)
            t = file(f).read()
            r.addrevision(t)
            new[f] = r.node(r.tip())

        # update manifest
        old = self.manifest.manifest(self.manifest.tip())
        old.update(new)
        for f in delete:
            del old[f]
        rev = self.manifest.addmanifest(old)

        # add changeset
        new = new.keys()
        new.sort()
        n = self.changelog.addchangeset(self.manifest.node(rev), new, "commit")
        self.current = n
        self.open("current", "w").write(str(self.current))

        if update: os.unlink(self.join("to-add"))
        if delete: os.unlink(self.join("to-delete"))

    def checkdir(self, path):
        d = os.path.dirname(path)
        if not d: return
        if not os.path.isdir(d):
            self.checkdir(d)
            os.mkdir(d)

    def checkout(self, rev):
        change = self.changelog.changeset(rev)
        mnode = change[0]
        mmap = self.manifest.manifest(self.manifest.rev(mnode))

        st = self.open("dircache", "w")
        l = mmap.keys()
        l.sort()
        for f in l:
            r = filelog(self, f)
            t = r.revision(r.rev(mmap[f]))
            try:
                file(f, "w").write(t)
            except:
                self.checkdir(f)
                file(f, "w").write(t)
                
            s = os.stat(f)
            e = (s.st_mode, s.st_size, s.st_mtime, len(f))
            e = struct.pack(">llll", *e)
            st.write(e + f)

        self.current = change
        self.open("current", "w").write(str(self.current))

    def diffdir(self, path):
        st = self.open("dircache").read()
        dc = {}
        pos = 0

        while pos < len(st):
            e = struct.unpack(">llll", st[pos:pos+16])
            l = e[3]
            pos += 16
            f = st[pos:pos + l]
            dc[f] = e[:3]
            pos += l

        changed = []
        added = []

        for dir, subdirs, files in os.walk(self.root):
            d = dir[len(self.root)+1:]
            if ".hg" in subdirs: subdirs.remove(".hg")
            
            for f in files:
                fn = os.path.join(d, f)
                s = os.stat(fn)
                if fn in dc:
                    c = dc[fn]
                    del dc[fn]
                    if c[1] != s.st_size:
                        changed.append(fn)
                        print "C", fn
                    elif c[0] != s.st_mode or c[2] != s.st_mtime:
                        t1 = file(fn).read()
                        t2 = self.file(fn).revision(self.current)
                        if t1 != t2:
                            changed.append(fn)
                            print "C", fn
                else:
                    added.append(f)
                    print "A", fn

        deleted = dc.keys()
        deleted.sort()
        for f in deleted:
            print "D", f

    def add(self, list):
        al = self.open("to-add", "a")
        st = self.open("dircache", "a")
        for f in list:
            al.write(f + "\n")
            s = os.stat(f)
            e = (s.st_mode, s.st_size, s.st_mtime, len(f))
            e = struct.pack(">llll", *e)
            st.write(e + f)

    def delete(self, list):
        dl = self.open("to-delete", "a")
        for f in list:
            dl.write(f + "\n")

    
